#include "Renderable.h"
#include "LayerManager.h"

int Renderable::lastID = 0;

Renderable::Renderable(int TypeID)
{
	m_width = 0;
	m_height = 0;
	
	m_bitmapBits = NULL;
	m_Texture = 0;

	ParentLayer = NULL;

	//set unique id for this session.
	id = ++lastID;
	m_typeID = TypeID;

	m_inverse = false;
}

Renderable::~Renderable(void)
{
	glDeleteTextures(1, &m_Texture);
	delete [] m_bitmapBits;
}

void Renderable::Initialize()
{
	m_layerManager = LayerManager::m_Manager;
	m_width = m_layerManager->m_width;
	m_height = m_layerManager->m_height;
	ResetCoords();

	m_variableCount = 0;
	SetVariables();

	delete[] m_bitmapBits;
	m_bitmapBits = NULL;
	
	//generate textures
	glGenTextures(1, &m_Texture);

	m_bitmapBits = new GLubyte[m_width*m_height];
}

void Renderable::MoveCoordinates(int delta_x, int delta_y)
{
	//xdiff, ydiff are in PIXELS
	//we want to translate from pixels to 3-space.

	float xdiff = delta_x, ydiff = delta_y;
	m_windowMapping.PixelDeltaToDelta(xdiff, ydiff);
	m_coords.x += xdiff;
	m_coords.y += ydiff;

	for(int i = 0; i < m_variableCount; i++)
	{
		m_layerVariables[i].MoveCoordinates(delta_x, delta_y);
	}

	UpdateWindowMapper();
}

//only update current layer.
void Renderable::UpdateWindowMapper()
{
	if(m_width > 0 && m_height > 0)
	{
		m_windowMapping.SetCoordinates(m_coords.y - m_zoomy/2, m_coords.x + m_zoomx/2, m_coords.x - m_zoomx/2, m_coords.y + m_zoomy/2);
		m_windowMapping.SetDestinationWindow(0, m_width, 0, m_height);
	}
}

//resets all logical coordinates, zoom, and cutoff.
void Renderable::ResetCoords()
{
	m_coords.x = 0;
	m_coords.y = 0;
	m_zoomx = 20;
	m_zoomy = 20;

	UpdateWindowMapper();
}

void Renderable::ChangeZoom(float zoom)
{
	zoom = MAX(0.1, zoom);
	assert(zoom > 0);

	m_zoomx = m_zoomy = zoom;
	UpdateWindowMapper();
}

void Renderable::IncrementZoom(float deltazoom)
{
	for(int i = 0; i < m_variableCount; i++)
	{
		m_layerVariables[i].IncrementZoom(deltazoom);
	}

	//one complete pass should reduce size of window by 1/2:
	float sizingfactor = 1 - (deltazoom);
	ChangeZoom(sizingfactor * m_zoomy);
}

void Renderable::SetCoordinates(float x, float y, float zoomX, float zoomY)
{
	assert(zoomX > 0);
	assert(zoomY > 0);
	m_zoomy = zoomY;
	m_zoomx = zoomX;
	m_coords.x = x;
	m_coords.y = y;

	UpdateWindowMapper();
}

void Renderable::AssignLayerToVariable(int variableIndex, Renderable* subLayer)
{
	if(variableIndex < m_variableCount)
	{
		//when assigning, set coordinates to where parent is.
		subLayer->SetCoordinates(m_coords.x, m_coords.y, m_zoomx, m_zoomy);
		//this next line links the sublayer's parent to this layer.
		m_layerVariables[variableIndex].SetRenderable(subLayer);
	}
}

void Renderable::ChangeParameter(int variableIndex, float newParameter)
{
	if(variableIndex < m_variableCount)
	{
		m_layerVariables[variableIndex].SetValue(newParameter);
	}
}

void Renderable::IncrementParameter(int variableIndex, float deltacutoff)
{
	ChangeParameter(variableIndex, m_layerVariables[variableIndex].GetOverrideValue() + deltacutoff);
}

void Renderable::UpdateTexture()
{
	// Create The Texture
	WriteBitmap();

	glBindTexture(GL_TEXTURE_2D, m_Texture);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);	// Linear Min Filter
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);	// Linear Mag Filter

	if(!m_bitmapBits)
	{
		//I don't think this part gets called - because we never recreate the bitmap.
		//resizing the window SHOULD do that, eventually.
		GLubyte gray = 0x33;
		glTexImage2D(GL_TEXTURE_2D, 0, 1, 1, 1, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, &gray);
	}
	else
		glTexImage2D(GL_TEXTURE_2D, 0, 1, m_width, m_height, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, m_bitmapBits);

}

void Renderable::Render(float zValue)
{
	glColor4f(1.0f,1.0f,1.0f, 1.0f);
	glBindTexture(GL_TEXTURE_2D, m_Texture);

	glBegin(GL_QUADS);
		// Front Face
		glTexCoord2f(0.0f, 0.0f); glVertex3f(0,		0,    zValue);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(1.0f,	0.0f, zValue);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(1.0f,  1.0f,  zValue);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(0.0f,  1.0f,  zValue);
	glEnd();

}

float Renderable::GetValueAt(int x, int y)
{
	//perform further standard checks
	float value = GetRawValueAt(x, y);
	if(m_inverse)
		return (1.0 - value);

	return value;
}

void Renderable::WriteBitmap()
{
	//creating 2d bitmap texture.
	GLubyte gray;

	for(int i = 0; i < m_width; i++)
	{
		for(int j = 0; j < m_height; j++)
		{
			gray = (int) 0xFF * GetValueAt(i, j);
			m_bitmapBits[i + m_width*j] = gray;
		}
	}
}